//
// Lighthouse3D.com CubeMap Sample
//
// Uses:
//  VS*L (www.lighthouse3d.com)
//
//
// This demo was built for learning purposes only.
// Some code could be severely optimised, but I tried to
// keep as simple and clear as possible.
//
// The code comes with no warranties, use it at your own risk.
// You may use it, or parts of it, wherever you want.
//
// If you do use it I would love to hear about it. Just post a comment
// at Lighthouse3D.com

// Have Fun :-)

#include <math.h>
#include <fstream>
#include <map>
#include <string>
#include <vector>


// include GLEW to access OpenGL 3.3 functions
#include <GL/glew.h>

// GLUT is the toolkit to interface with the OS
#include <GL/freeglut.h>

// Use Very Simple Libs
#include "../VSL/vslibs.h"

GLuint fbo, texFBO[2];


VSMathLib *vsml;
VSShaderLib shader, cmShader;
VSResModelLib myModel1, myModel2, myModel3, myModel4, 
		myModel5, myModel6, myModel7, myModel8, myModel9;



// Mouse Tracking Variables
int startX, startY, tracking = 0;

// Camera Position
float camX, camY, camZ;
// Camera Spherical Coordinates
float alpha = 0.0f, beta = 0.0f;
float r = 5.0f;


// Frame counting and FPS computation
long myTime,timebase = 0,frame = 0;
char s[32];

float a= 0.0f;


// ------------------------------------------------------------
//
// Reshape Callback Function
//

void changeSize(int w, int h) {

	float ratio;
	// Prevent a divide by zero, when window is too short
	// (you cant make a window of zero width).
	if(h == 0)
		h = 1;

	// Set the viewport to be the entire window
    glViewport(0, 0, w, h);

	// Set the projection matrix
	ratio = (1.0f * w) / h;
	vsml->loadIdentity(VSMathLib::PROJECTION);
	vsml->perspective(53.13f, ratio, 0.1f, 10000.0f);
}


// ------------------------------------------------------------
//
// Render stuff
//

void renderScene(void) {

	// Render to FBO

	// Bind FPO
	glBindFramebuffer(GL_FRAMEBUFFER, fbo);

	// Set Drawing buffers
	GLuint attachments[2] = {GL_COLOR_ATTACHMENT0, GL_COLOR_ATTACHMENT1};
	glDrawBuffers(2,  attachments);

	// this is the resolution of the render target
	changeSize(400,400);

	// clear color and depth
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	// Load identity matrices
	vsml->loadIdentity(VSMathLib::VIEW);
	vsml->loadIdentity(VSMathLib::MODEL);

	// set camera
	vsml->lookAt(camX, camY, camZ, 0,0,0, 0,1,0);

	// pass camera position to the shader
	float c[3];
	c[0] = camX; c[1] = camY; c[2] = camZ;
	cmShader.setUniform("camPosWorld", c);

	// use regular shader
	glUseProgram(shader.getProgramIndex());

	// render the six planes of the skybox
	vsml->pushMatrix(VSMathLib::MODEL);
	vsml->translate(0.0, -4.0, 0.0);
	vsml->scale(4.0, 4.0, 4.0);
	myModel2.render();
	vsml->popMatrix(VSMathLib::MODEL);

	vsml->pushMatrix(VSMathLib::MODEL);
	vsml->translate(0.0, 4.0, 0.0);
	vsml->scale(4.0, 4.0, 4.0);
	vsml->rotate(180, 1.0, 0.0, 0.0);
	myModel5.render();
	vsml->popMatrix(VSMathLib::MODEL);

	vsml->pushMatrix(VSMathLib::MODEL);
	vsml->translate(0.0, 0.0, -4.0);
	vsml->scale(4.0, 4.0, 4.0);
	vsml->rotate(90, 1.0, 0.0, 0.0);
	vsml->rotate(180, 0.0, 1.0, 0.0);
	myModel3.render();
	vsml->popMatrix(VSMathLib::MODEL);

	vsml->pushMatrix(VSMathLib::MODEL);
	vsml->translate(0.0, 0.0, 4.0);
	vsml->scale(4.0, 4.0, 4.0);
	vsml->rotate(-90, 1.0, 0.0, 0.0);
	myModel4.render();
	vsml->popMatrix(VSMathLib::MODEL);

	vsml->pushMatrix(VSMathLib::MODEL);
	vsml->translate(4.0, 0.0, 0.0);
	vsml->scale(4.0, 4.0, 4.0);
	vsml->rotate(90, 0.0, 0.0, 1.0);
	vsml->rotate(90, 0.0, 1.0, 0.0);
	myModel6.render();
	vsml->popMatrix(VSMathLib::MODEL);

	vsml->pushMatrix(VSMathLib::MODEL);
	vsml->translate(-4.0, 0.0, 0.0);
	vsml->scale(4.0, 4.0, 4.0);
	vsml->rotate(-90, 0.0, 0.0, 1.0);
	vsml->rotate(-90, 0.0, 1.0, 0.0);
	myModel7.render();
	vsml->popMatrix(VSMathLib::MODEL);

	// use the cubemap shader
	glUseProgram(cmShader.getProgramIndex());
	// render model
	vsml->pushMatrix(VSMathLib::MODEL);
	vsml->rotate(a, 0.0, 1.0, 0.0);
	myModel1.render();
	vsml->popMatrix(VSMathLib::MODEL);
	a+=0.01f;


	// turn off FBO
	glBindFramebuffer(GL_FRAMEBUFFER, 0);
	glUseProgram(shader.getProgramIndex());

	// Draw plane with texture
	// this will be loaded with all matrices set to identity

	glViewport(0,0,800,400);
	// clear default frame buffer
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	// load identity matrices
	vsml->loadIdentity(VSMathLib::MODEL);
	vsml->loadIdentity(VSMathLib::VIEW);
	vsml->loadIdentity(VSMathLib::PROJECTION);

	// render left image
	glViewport(0,0,400,400);
	// render plane
	vsml->pushMatrix(VSMathLib::MODEL);
	vsml->rotate(90, 1.0, 0.0, 0.0);
	myModel8.render();
	vsml->popMatrix(VSMathLib::MODEL);

	// render right image
	glViewport(400,0,400,400);
	// render plane
	vsml->pushMatrix(VSMathLib::MODEL);
	vsml->rotate(90, 1.0, 0.0, 0.0);
	myModel9.render();
	vsml->popMatrix(VSMathLib::MODEL);

					
	// FPS computation and display
	frame++;
	myTime=glutGet(GLUT_ELAPSED_TIME);
	if (myTime - timebase > 1000) {
			sprintf(s,"FPS:%4.2f",
				frame*1000.0/(myTime-timebase) );
		timebase = myTime;
		frame = 0;
		glutSetWindowTitle(s);
	}

	// restore viewport
	glViewport(0,0,800,400);
	glutSwapBuffers();
}


// ------------------------------------------------------------
//
// Events from the Keyboard
//

void processKeys(unsigned char key, int xx, int yy)
{
	switch(key) {

		case 27:

			glutLeaveMainLoop();
			break;

		case 'm': glEnable(GL_MULTISAMPLE); break;
		case 'n': glDisable(GL_MULTISAMPLE); break;
				break;

	}

	//  uncomment this if not using an idle func
//	glutPostRedisplay();
}


// ------------------------------------------------------------
//
// Mouse Events
//

void processMouseButtons(int button, int state, int xx, int yy)
{
	// start tracking the mouse
	if (state == GLUT_DOWN)  {
		startX = xx;
		startY = yy;
		if (button == GLUT_LEFT_BUTTON)
			tracking = 1;
		else if (button == GLUT_RIGHT_BUTTON)
			tracking = 2;
	}

	//stop tracking the mouse
	else if (state == GLUT_UP) {
		if (tracking == 1) {
			alpha -= (xx - startX);
			beta += (yy - startY);
		}
		else if (tracking == 2) {
			r += (yy - startY) * 0.01f;
			if (r < 0.1f)
				r = 0.1f;
		}
		tracking = 0;
	}
}

// Track mouse motion while buttons are pressed

void processMouseMotion(int xx, int yy)
{

	int deltaX, deltaY;
	float alphaAux, betaAux;
	float rAux;

	deltaX =  - xx + startX;
	deltaY =    yy - startY;

	// left mouse button: move camera
	if (tracking == 1) {


		alphaAux = alpha + deltaX;
		betaAux = beta + deltaY;

		if (betaAux > 85.0f)
			betaAux = 85.0f;
		else if (betaAux < -85.0f)
			betaAux = -85.0f;
		rAux = r;
	}
	// right mouse button: zoom
	else if (tracking == 2) {

		alphaAux = alpha;
		betaAux = beta;
		rAux = r + (deltaY * 0.01f);
		if (rAux < 0.1f)
			rAux = 0.1f;
	}

	camX = rAux * sin(alphaAux * 3.14f / 180.0f) * cos(betaAux * 3.14f / 180.0f);
	camZ = rAux * cos(alphaAux * 3.14f / 180.0f) * cos(betaAux * 3.14f / 180.0f);
	camY = rAux *   						       sin(betaAux * 3.14f / 180.0f);

//  uncomment this if not using an idle func
//	glutPostRedisplay();
}





void mouseWheel(int wheel, int direction, int x, int y) {

	r += direction * 0.1f;
	if (r < 0.1f)
		r = 0.1f;

	camX = r * sin(alpha * 3.14f / 180.0f) * cos(beta * 3.14f / 180.0f);
	camZ = r * cos(alpha * 3.14f / 180.0f) * cos(beta * 3.14f / 180.0f);
	camY = r *   						     sin(beta * 3.14f / 180.0f);

//  uncomment this if not using an idle func
//	glutPostRedisplay();
}



// --------------------------------------------------------
//
// Shader Stuff
//


GLuint setupShaders() {

	// Shader for fonts and models
	shader.init();
	shader.loadShader(VSShaderLib::VERTEX_SHADER, "shaders/color.vert");
	shader.loadShader(VSShaderLib::FRAGMENT_SHADER, "shaders/color.frag");

	// set semantics for the shader variables
	shader.setProgramOutput(0,"outputF");
	shader.setVertexAttribName(VSShaderLib::VERTEX_COORD_ATTRIB, "position");
	shader.setVertexAttribName(VSShaderLib::TEXTURE_COORD_ATTRIB, "texCoord");

	shader.prepareProgram();

	VSGLInfoLib::getUniformsInfo(shader.getProgramIndex());

	printf("InfoLog\n%s\n\n", shader.getAllInfoLogs().c_str());

	// add sampler uniforms
	shader.setUniform("texUnit", 0);




	// cube map shader
	cmShader.init();
	cmShader.loadShader(VSShaderLib::VERTEX_SHADER, "shaders/cubemap.vert");
	cmShader.loadShader(VSShaderLib::FRAGMENT_SHADER, "shaders/cubemap.frag");

	cmShader.setProgramOutput(0,"color1");
	cmShader.setProgramOutput(1,"color2");

	cmShader.setVertexAttribName(VSShaderLib::VERTEX_COORD_ATTRIB, "position");
	cmShader.setVertexAttribName(VSShaderLib::NORMAL_ATTRIB, "normal");

	cmShader.prepareProgram();

	printf("InfoLog for Cube Map Shader\n%s\n\n", cmShader.getAllInfoLogs().c_str());

	cmShader.setUniform("texUnit",4);

	VSGLInfoLib::getUniformsInfo(cmShader.getProgramIndex());


	return(shader.isProgramValid() && cmShader.isProgramValid());
}


// ------------------------------------------------------------
//
// Model loading and OpenGL setup
//

// This function creates a RGBA texture with the given dimensions and
// returns the texture id

GLuint createRGBATexture(int w, int h) {

	GLuint tex;

	glGenTextures(1, &tex);

	glBindTexture(GL_TEXTURE_2D, tex);

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR); 
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, 
						w,
						h, 
						0, GL_RGBA, GL_UNSIGNED_BYTE,
						NULL); 

	glBindTexture(GL_TEXTURE_2D, 0);

	return(tex);
}


bool prepareFBO(int w, int h, int colorCount) {

	// Generate one frame buffer
	glGenFramebuffers(1, &fbo);
	// bind it
	glBindFramebuffer(GL_DRAW_FRAMEBUFFER, fbo);
	// attach textures for colors
	for (int i = 0; i < colorCount; ++i) {
	
		texFBO[i] = createRGBATexture(w,h);
		glFramebufferTexture(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0+i, texFBO[i], 0);
	}

	// attach renderbuffer for depth
	GLuint fb;
	glGenRenderbuffers(1, &fb);
	glBindRenderbuffer(GL_RENDERBUFFER, fb);
	glRenderbufferStorage(GL_RENDERBUFFER, GL_DEPTH_COMPONENT24, w,h);
	glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_RENDERBUFFER, fb);

	// check if everything is OK
	GLenum e = glCheckFramebufferStatus(GL_DRAW_FRAMEBUFFER);
	bool res = false;
	switch (e) {
	
		case GL_FRAMEBUFFER_UNDEFINED:
			printf("FBO Undefined\n");
			break;
		case GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT :
			printf("FBO Incomplete Attachment\n");
			break;
		case GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT :
			printf("FBO Missing Attachment\n");
			break;
		case GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER :
			printf("FBO Incomplete Draw Buffer\n");
			break;
		case GL_FRAMEBUFFER_UNSUPPORTED :
			printf("FBO Unsupported\n");
			break;
		case GL_FRAMEBUFFER_COMPLETE:
			res = true;
			printf("FBO OK\n");
			break;
		default:
			printf("FBO Problem?\n");
	}

	// unbind fbo
	glBindFramebuffer(GL_DRAW_FRAMEBUFFER, 0);

	return res;
}


int init()
{
	// load models
	if (myModel1.load("../../models/teapot.3ds") &&
		myModel2.load("../../models/plane.obj")) {


		myModel3.clone(&myModel2);
		myModel4.clone(&myModel2);
		myModel5.clone(&myModel2);
		myModel6.clone(&myModel2);
		myModel7.clone(&myModel2);
		myModel8.clone(&myModel2);
		myModel9.clone(&myModel2);

		myModel1.addCubeMapTexture(4, 
			"../../textures/CM/CloudyHills_posx_512.png", 
			"../../textures/CM/CloudyHills_negx_512.png", 
			"../../textures/CM/CloudyHills_posy_512.png", 
			"../../textures/CM/CloudyHills_negy_512.png", 
			"../../textures/CM/CloudyHills_posz_512.png", 
			"../../textures/CM/CloudyHills_negz_512.png");

		printf("Models loaded\n%s", myModel1.getInfo().c_str());

		myModel2.addTexture(0, "../../textures/CM/CloudyHills_negy_512.png");
		myModel3.addTexture(0, "../../textures/CM/CloudyHills_negz_512.png");
		myModel4.addTexture(0, "../../textures/CM/CloudyHills_posz_512.png");
		myModel5.addTexture(0, "../../textures/CM/CloudyHills_posy_512.png");
		myModel6.addTexture(0, "../../textures/CM/CloudyHills_posx_512.png");
		myModel7.addTexture(0, "../../textures/CM/CloudyHills_negx_512.png");

		// some GL settings
		glEnable(GL_DEPTH_TEST);
		glEnable(GL_CULL_FACE);
		glEnable(GL_MULTISAMPLE);
		glClearColor(0.5f, 0.5f, 0.5f, 1.0f);

		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

		if (!prepareFBO(400,400,2))
			return false;

		myModel8.setTexture(0,texFBO[0]);
		myModel9.setTexture(0,texFBO[1]);

		camX = r * sin(alpha * 3.14f / 180.0f) * cos(beta * 3.14f / 180.0f);
		camZ = r * cos(alpha * 3.14f / 180.0f) * cos(beta * 3.14f / 180.0f);
		camY = r *   						     sin(beta * 3.14f / 180.0f);


		return true;
	}
	else {
		printf("%s\n",myModel1.getErrors().c_str());
		return false;
	}
}


void initVSL() {

	VSResourceLib::setMaterialBlockName("Material");

//	 Init VSML
	vsml = VSMathLib::getInstance();
	vsml->setUniformBlockName("Matrices");
	vsml->setUniformName(VSMathLib::PROJ_VIEW_MODEL, "m_pvm");
	vsml->setUniformName(VSMathLib::NORMAL_MODEL, "m_normalModel");
	vsml->setUniformName(VSMathLib::MODEL, "m_model");
}


// ------------------------------------------------------------
//
// Main function
//


int main(int argc, char **argv) {

//  GLUT initialization
	glutInit(&argc, argv);

	glutInitDisplayMode(GLUT_DEPTH|GLUT_DOUBLE|GLUT_RGBA|GLUT_MULTISAMPLE);

	glutInitContextVersion (3, 3);
	glutInitContextProfile (GLUT_CORE_PROFILE );
	glutInitContextFlags(GLUT_DEBUG);

	glutInitWindowPosition(100,100);
	glutInitWindowSize(800,400);
	glutCreateWindow("Lighthouse3D - Cube Map Demo");


//  Callback Registration
	glutDisplayFunc(renderScene);
	glutReshapeFunc(changeSize);
	glutIdleFunc(renderScene);

//	Mouse and Keyboard Callbacks
	glutKeyboardFunc(processKeys);
	glutMouseFunc(processMouseButtons);
	glutMotionFunc(processMouseMotion);

	glutMouseWheelFunc ( mouseWheel ) ;

//	return from main loop
	glutSetOption(GLUT_ACTION_ON_WINDOW_CLOSE, GLUT_ACTION_GLUTMAINLOOP_RETURNS);

//	Init GLEW
	glewExperimental = GL_TRUE;
	glewInit();

   printf ("Vendor: %s\n", glGetString (GL_VENDOR));
   printf ("Renderer: %s\n", glGetString (GL_RENDERER));
   printf ("Version: %s\n", glGetString (GL_VERSION));
   printf ("GLSL: %s\n", glGetString (GL_SHADING_LANGUAGE_VERSION));
   int param;
   glGetIntegerv(GL_CONTEXT_PROFILE_MASK, &param);
   if (param == GL_CONTEXT_CORE_PROFILE_BIT)
	   printf("Context Profile: Core\n");
   else
	   printf("Context Profile: Compatibility\n");

	if (!setupShaders())
		exit(1);

	if (!init()) {
		exit(1);
	}

	initVSL();

	//  GLUT main loop
	glutMainLoop();

	return(1);

}

